import os
import yaml

#
# Set literal style to block (more readable for long strings)
#
def selective_representer(dumper, data):
    return dumper.represent_scalar(u"tag:yaml.org,2002:str", data,
                                   style="|" if "\n" in data else None)

yaml.add_representer(str, selective_representer)


def get_config(file_name):

    config = {}

    path = os.path.expanduser(file_name)
    if os.path.exists(path):
        with open(path) as fp:
            config = yaml.load(fp, Loader=yaml.FullLoader)

    # defines sections in dict for writing
    if 'compute' not in config:
        config['compute'] = {'configurations': {}, 'settings': {}}
    if 'storage' not in config:
        config['storage'] = {'configurations': {}, 'settings': {}}

    return config

def write_config(config, file_name):

    # Removes empty sections
    if len(config.get('storage', {}).get('configurations', {})) == 0:
        del config['storage']
    if len(config.get('compute', {}).get('configurations', {})) == 0:
        del config['compute']

    path = os.path.expanduser(file_name)
    # print ("PATH", path)
    if not os.path.exists(path):
        if path.count('/') >= 2:
            os.mkdir(path[:path.rfind('/')])
    with open(path, 'w') as fp:
        yaml.dump(config, fp)




def compute(config):
    print('\nThe following compute services are configured:')
    if 'compute' in config and len(config['compute'].get('configurations', [])) > 0:
        for service in config['compute']['configurations']:
            print(f'  * {service}')
    else:
        print('(none)')

def storage(config):

    print('\nThe following storage services are configured:')
    if 'storage' in config and len(config['storage'].get('configurations', [])) > 0:
        for service in config['storage']['configurations']:
            print(f'  * {service}')
    else:
        print('(none)')

def all(config):

    compute(config)
    storage(config)



#!/usr/bin/env python3
import os, sys, argparse, time, json, getpass
#
# the BDFL does not admire scripts which are also importable modules
# well, frack him -- this is how we roll
#
#so absolute imports work in script mode, we need to import from the parent folder
opath = os.path.abspath(".")
abspath = os.path.abspath(__file__)
# print ("CLI PATHS:", opath, abspath)
abspath = abspath[:abspath.rfind('/') + 1]
os.chdir(abspath)
abspath = os.path.abspath("..")
sys.path.insert(0, abspath)

os.chdir(opath)

from burst.burst import *

#
# typing reduction act
#
def complete(x, a):
    num = 0
    match = []
    for k in a:
        if k[:len(x)] == x:
            num += 1
            match.append(k)
    return match, num

actions = {
    # None,
    'build':            "burst build                                    |build project",
    'run':              "burst run <command>                            |run <command> on remote server",
    'help':             "burst help                                     |print helpful information",
    'list-servers':     "burst list-servers                             |list available servers; display time till automatic stop",
    'status':           "burst status                                   |show status of remote task (if running)",
    'stop-server':      "burst stop-server                              |force-stop server (prompts for confirmation)",
    'terminate-server': "burst termimate-server                         |terminate (delete) remote server (prompts for confirmation)",
    'attach':           "burst attach                                   |attach stdin, stdout, stderr to background process. ctl-C detaches",
    'sync':             "burst sync                                     |synchronize local directory to remote",
    'kill':             "burst kill                                     |stop docker process on remote",
    'actions':          "burst actions                                  |list available actions",
    'configure':        "burst actions                                  |Interactive configuration",
}

actions_keys_sorted = list(actions)
actions_keys_sorted.sort()

#
# This hack ensures we do not collect new, undocumented 'actions' (subcommands)
#
def switch(action, *args):
    # print ("SWITCH:", action, args, actions, action in actions)
    if action == None:
        return False
    if action not in actions:
        raise Exception("Unknown action: %s  try: 'burst help'" % action)
    for a in args:
        if a == action:
            return True
    return False


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, add_help=False)
    add = parser.add_argument
    add("action", nargs='?',                                                                help="type 'burst actions' to list available actions")
    add("--background", "-b",   action="store_true",                                        help="Run task in background mode")
    add("--compute-access",     metavar="KEY", dest='access',                               help="libcloud username (aws: ACCESS_KEY)")
    add("--compute-provider",   dest='provider',default='EC2',                              help="GCE, EC2 etc.")
    add("--compute-region",     dest='region',                                              help="libcloud location (aws: region)")
    add("--compute-secret",     dest='secret',                                              help="libcloud password (aws: SECRET)")
    add("--compute-service",    dest='compute_config',metavar="COMPUTE_SERVICE",            help="override default compute configuration service")
    add("--config-file",        metavar="FILE", dest='configfile',                          help="override default config.yml")
    add("--docker-file",        dest='dockerfile', type=str, default="Dockerfile", metavar="FILE",
                                                                                            help="Docker file (defaults to ./Dockerfile)")
    add("--docker-port",        dest='dockerdport', type=int, default=2377, metavar="PORT", help="local port to map to remote host docker daemon"
                                                                                                 "(default: 2377)")
    add("--gcs-project",        dest='project',                                             help="Google Cloud project ID")
    add("--gpu",                action="store_true",                                        help="Build with gpu")
    add("--help",               action="store_true",                                        help="Print usage info")
    add("--local",              action="store_true",                                        help="run on local device")
    add("--no-gpu",             action="store_true",                                        help="Build without gpu")
    add("--pubkey-file",        dest='pubkey',                                              help="public key to access server (defaults to ~/.ssh/id_rsa.pub)")
    add("--session-name",       metavar="NAME", dest='burst_user',                          help="Burst session name (defaults to burst-username; "
                                                                                                 "different sessions launch new machine instances)")
    add("--stop",               type=int, default=900, metavar="SECONDS",                   help="seconds before server is stopped (default 900) "
                                                                                                 "0 means never. Use action 'stop' to force stop")
    add("--storage-mount",      dest="cloudmap", type=str, default="",  metavar="STORAGE:MOUNT",
                                                                                            help="map (mount) burst storage service to local folder")
    add("--storage-service",    dest="storage_config", metavar="STORAGE_SERVICE",           help="override default storage configuration")
    add("--tunnel-port", "-p",  dest='portmap', action="append", metavar="LOCAL[:REMOTE]",  help="port mapping; example: -p 8080 or -p 8081:8080")
    add("--verbose", "-v",      dest='verbosity', type=int, default=0,                      help="-1: just task output 0: status 1-127: more verbose"
                                                                                                 "(default: -1)")
    add("--version",            action="store_true",                                        help="Print version # & exit")
    add("--vm-image",           dest='image',                                               help="libcloud image (aws: ami image_id)")
    add("--vm-type",            dest='size',                                                help="aws: instance_type; gce: size)")
    add("--vm-username",        dest='sshuser', default="ubuntu",                           help="remote server username for login")

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(1)
    #
    # this got a bit tricky.
    # we want to parse args BEFORE the main command as burst options
    # and pass all args AFTER the main command to the command when it runs remotely
    #
    argv = sys.argv[1:]
    try:
        args, task_args = parser.parse_known_args(argv)
    except SystemExit:
        traceback.print_exc()
        print ("There was an error parsing arguments. If there is an argument conflict, try 'run -- yourcommands'")
        exit()
    set_verbosity(args.verbosity)

    if args.action == None:
        action = None
    else:
        action, matches = complete(args.action, actions)
        if matches > 1:
            raise Exception(f"Ambiguous action: {args.action} could be one of: {', '.join(action)}")
        elif matches == 0:
            raise Exception("Unknown action '%s'; try: 'burst --help'" % args.action)
        else:
            action = action[0]
            vvprint (f"Expanding action: {args.action} --> {action}")

    vvprint ("ARGV:", argv)
    vvprint ("BURST:")
    for k, v in args.__dict__.items():
        if v:
            vvprint (f"  {k}=={v}")
    vvprint ("TASK:")
    for k in task_args:
        vvprint (" ", k)

    if action == 'build' and args.verbosity < 1:
        set_verbosity(9)

    if args.help:
        parser.print_help()
        sys.exit(1)

    #override config credentials on command line: --access implies all must be provided
    if args.access:
        args_compute = dictobj()
        args_compute.access = args.access
        args_compute.secret = args.secret
        args_compute.region = args.region
        args_compute.project = args.project
        args_compute.provider = args.provider
    else:
        burst_conf = {}

        #command line overrides:
        if args.compute_config:
            burst_conf['compute_config'] = args.compute_config

        if args.storage_config:
            burst_conf['storage_config'] = args.storage_config

        if args.configfile:
            burst_conf['configfile'] = args.configfile

    if args.local:
        vprint (args)
        parser.error("when specifying --local, do not set --vm-username or --session-name")
        exit()
    # t0 = time.time()
    # while time.time()-t0 < args.delay:
    #     vprint ("%d seconds till action" % (args.delay+.5+t0-time.time()))
    #     time.sleep(5)

    #set default burst_user if necessary:
    if not (args.burst_user or args.local or args.version):
        burst_user = getpass.getuser()
        args.burst_user = "burst-" + burst_user
        vprint ("Session: %s" % args.burst_user)

    if action != 'run' and len(task_args) > 0:
        raise Exception(f"Unknown arguments: {task_args}")

############################################################################
    # #master switch clause. First, stand-alone options
    if switch(action, 'help'):
        # print ("DBG:", action, args.action, argv)
        print ("type 'burst --help' for help, 'burst actions' for documentation on available actions")
        exit()

    elif switch(action, 'actions'):
        # print ("DBG:", action, args.action, argv)
        print (" " * 80 + "\r")
        print ("Available actions (note you can abbreviate if unambiguous, 'burst act'):")
        for act in actions_keys_sorted:
            print (f"  {actions[act]}")
        exit()

    elif switch(action, 'list-servers'):
        init(burst_conf)
        # pprint(get_config())
        cconf = get_config()['compute_config']
        v0print ("-------------------------------------------------------------\nSessions with config %s & user %s:" % (cconf, args.burst_user))
        for n, s in list_servers(args.burst_user, burst_conf):
            # print ("DBG:", n.public_ips[0])
            print (s)
            if n.state.lower()=='running':
                cmd = f"ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error ubuntu@{n.public_ips[0]} 'screen -r -md -X hardcopy .burst_monitor.log; tail -n 2 .burst_monitor.log'"
                os.system(cmd)
        v0print ("-------------------------------------------------------------")

    elif switch(action, 'stop-server'):
        v0print ("-------------------------------------------------------------")
        count = 0
        for node, s in list_servers(args.burst_user, burst_conf):
            if node.state == "stopped":
                continue
            count += 1
            yes = input("Stopping (warm shutdown) %s, are you sure? (y/n)" % s)
            if yes=='y':
                stop_server(node)
            else:
                print ("Aborted")
        if not count:
            print ("no servers to shut down")
        v0print ("-------------------------------------------------------------")

    elif switch(action, 'terminate-server'):
        v0print ("-------------------------------------------------------------")
        count = 0
        for node, s in list_servers(args.burst_user, burst_conf, terminated=False):
            count += 1
            yes = input("Terminating %s, are you sure? (y/n)" % s)
            if yes=='y':
                os.system("rm .burst-gpu")
                terminate_server(node)
            else:
                print ("Aborted")
        if not count:
            print ("no servers to terminate")
        v0print ("-------------------------------------------------------------")

    elif switch(action, 'attach'):
        tunnel = None
        init(burst_conf)
        cconf = get_config()['compute_config']
        url = None
        for node, s in list_servers(args.burst_user, burst_conf):
            vvprint (node, s)
            if node.state.upper() == 'RUNNING':
                if url:
                    raise Exception("multiple docker processes running, this is not supported")
                url = node.public_ips[0]
                break
        if not url:
            print ("No process running")
        else:
            vvprint (f"Attaching to docker process on {url}")
            tunnel, _ = ssh_tunnel(url, args.sshuser, args.portmap, args.dockerdport)
            vvprint ("Tunnel:", tunnel)
            cmd = ["docker", "-H localhost:%s" % args.dockerdport, "ps", "--format", '{{json .}}']
            vvprint (cmd)
            out, err = run(cmd)
            vvprint("PS returns:", out)
            if not out:
                print ("\nNo Docker process found")
            else:
                try:
                    did = json.loads(out)
                    v0print ("Attaching to docker process", did['ID'])
                    cmd = f"docker -H localhost:{args.dockerdport} attach --sig-proxy=false {did['ID']}"
                    vvprint (cmd)
                    vprint("ctrl-C only detaches;  'burst kill' to stop")
                    v0print ("---------------------OUTPUT-----------------------")
                    os.system(cmd)
                    v0print ("----------------------END-------------------------")
                except:
                    print ("\nFailed to attach:", out)
                    sys.stdout.flush()
        if tunnel:
            tunnel.kill()

    elif switch(action, 'kill'):
        tunnel = None
        init(burst_conf)
        cconf = get_config()['compute_config']
        url = None
        for node, s in list_servers(args.burst_user, burst_conf):
            vvprint (node, s)
            if node.state.upper() == 'RUNNING':
                if url:
                    raise Exception("multiple docker processes running, this is not supported")
                url = node.public_ips[0]
                break
        if not url:
            print ("No process running")
        else:
            vvprint (f"Terminating Docker process on {url}")
            tunnel, _ = ssh_tunnel(url, args.sshuser, args.portmap, args.dockerdport)
            vvprint ("Tunnel:", tunnel)
            cmd = ["docker", "-H localhost:%s" % args.dockerdport, "ps", "--format", '{{json .}}']
            vvprint (cmd)
            out, err = run(cmd)
            vvprint("PS returns:", out)
            if not out:
                print ("\nNo Docker process found")
            else:
                try:
                    did = json.loads(out)
                    yes = input(f"Terminating Docker process {did['ID']}, are you sure? (y/n)")
                    if yes == 'y':
                        cmd = f"docker -H localhost:{args.dockerdport} stop {did['ID']}"
                        vvprint (cmd)
                        os.system(cmd)
                        print ("Process terminated")
                    else:
                        print("Aborted")
                except:
                    print ("\nError:", out)
                    sys.stdout.flush()
        if tunnel:
            tunnel.kill()

    elif switch(action, 'status'):
        tunnel = None
        init(burst_conf)
        cconf = get_config()['compute_config']
        url = None
        for node, s in list_servers(args.burst_user, burst_conf):
            vvprint (node, s)
            if node.state.upper() == 'RUNNING':
                if url:
                    raise Exception("multiple docker processes running, this is not supported")
                url = node.public_ips[0]
                break
        if not url:
            v0print("-------------------------------------------------------------")
            print ("No remote host running")
            v0print("-------------------------------------------------------------")
        else:
            vvprint (f"Looking for docker process on {url}")
            tunnel, _ = ssh_tunnel(url, args.sshuser, args.portmap, args.dockerdport)
            vvprint ("Tunnel:", tunnel)
            cmd = ["docker", "-H localhost:%s" % args.dockerdport, "ps", "--no-trunc", "--format", '{{json .}}']
            vvprint (cmd)
            out, err = run(cmd)
            vvprint("PS returns:", out)
            if not out:
                print ("\nNo Docker process found")
            else:
                try:
                    did = json.loads(out)
                    v0print("-------------------------------------------------------------")
                    print (f"Docker process ID: {did['ID'][:12]}\n"
                           f"Status: {did['Status']}\n"
                           f"Command: {did['Command']}")
                           # f"Mounts: {did['Mounts']}")
                    v0print("-------------------------------------------------------------")
                except:
                    print ("\nError:", out)
                    sys.stdout.flush()
        if tunnel:
            tunnel.kill()

    elif args.version:
        print ("VERSION:", version)

    elif switch(action, 'configure'):
        if args.configfile:
            yam = args.configfile
        else:
            yam = os.environ['HOME'] + "/.burst/config.yml"
        os.system("burst-config --config_path %s" % yam)

    elif switch(action, 'build', 'run', 'sync'):
        #no stand-alone options; do burst for reals
        if args.local:
            pubkey = None
        else:
            if args.pubkey==None:
                try:
                    f=open(os.path.expanduser("~") + "/.ssh/id_rsa.pub")             #FIXME: a bit cheeky
                    pubkey=f.read()
                    f.close()
                except:
                    print ("Public key not found in usual place; please specify --pubkey")

        # args_gpus = args.gpus
        # if args.gpus == None:
        #     if os.path.exists(".burst_gpus"):
        #         args_gpus = open(".burst_gpus").read().strip()
        #     else:
        #         raise Exception("no .burst_gpus; specify --gpus")
        # if args_gpus.lower() != 'none':
        #     if args.size == None:
        #         size = 'DEFAULT_GPU_SIZE'
        #     else:
        #         size = args.size
        #     if args.image == None:
        #         image = 'DEFAULT_GPU_IMAGE'
        #     else:
        #         image = args.image
        # else:
        #     if args.size == None:
        #         size = 'DEFAULT_SIZE'
        #     else:
        #         size = args.size
        #     if args.image == None:
        #         image = 'DEFAULT_IMAGE'
        #     else:
        #         image = args.image

        #if we are launching, need to know gpu
        if not os.path.exists(".burst-gpu"):
            if not (args.gpu or args.no_gpu):
                raise Exception("Must specify --gpu or --no-gpu for initial build")
            f = open(".burst-gpu", 'w')
            f.write(f"{args.gpu}")
            f.close()
        f = open(".burst-gpu")
        gpu = f.read().strip().lower()=='true'
        f.close()

        #sanity clause
        if (gpu and args.no_gpu) or ((not gpu) and args.gpu):
            raise Exception("Gpu status can only be changed with fresh launch (terminate & rebuild)")

        #blech
        if gpu:
            if args.size == None:
                size = 'DEFAULT_GPU_SIZE'
            else:
                size = args.size
            if args.image == None:
                image = 'DEFAULT_GPU_IMAGE'
            else:
                image = args.image
        else:
            if args.size == None:
                size = 'DEFAULT_SIZE'
            else:
                size = args.size
            if args.image == None:
                image = 'DEFAULT_IMAGE'
            else:
                image = args.image

        if action == 'build':
            task_args = ['echo', 'Build phase 1 success']

        #let's do this thing
        error = burst(task_args, sshuser=args.sshuser,
              burst_user=args.burst_user, gpu=gpu, ports=args.portmap, stop=args.stop,
              image=image, size=size, pubkey=pubkey, dockerfile=args.dockerfile, cloudmap=args.cloudmap,
              dockerdport=args.dockerdport, bgd = args.background, sync_only = action=='sync', conf = burst_conf)

        if error:
            v0print ("Build failed")
        else:
            if action == 'build':
                v0print()
                print ("Build phase 2 success")

            vprint ("DONE")
            v0print()
    else:
        vprint()
        print ("Unknown action:", action)



import sys

VERBOSITY = 0
LIMIT = 100

def set_verbosity(v):
    global VERBOSITY
    VERBOSITY = v

def get_verbosity():
    return VERBOSITY

def _vprint_nolf(*args, **kw):
    # return
    if 'file' not in kw:
        file = sys.stdout
    sarg = ""
    for arg in args:
        sarg += (str(arg) + " ").rstrip()
    s =sarg[:LIMIT]
    s += " " * (LIMIT - len(s))
    s = s.replace("\n", "\r")
    s += "\r"
    file.write("burst: %s" % s)

def v0print(*args, **kw):
    if VERBOSITY >=0:
        print(*args, **kw)

def vprint(*args, **kw):
    if VERBOSITY == 0:
        _vprint_nolf(*args, **kw)
    elif VERBOSITY & 1:
        print (*args, **kw)

def vvprint(*args, **kw):
    if VERBOSITY & 2:
        print (*args, **kw)

#docker verbosity
def get_piper():
    if VERBOSITY < 3:
        return ">/dev/null 2>/dev/null"
    elif VERBOSITY & 8:
        return ""
    elif VERBOSITY & 4:
        return ">/dev/null"
    return ""

def get_dockrunflags():
    if VERBOSITY & 128:
        return "-ti"            #run in foreground
    return "-d"

#rsync verbosity
def get_rsync_v():
    if VERBOSITY & 16:
        return "v --progress"
    if VERBOSITY & 32:
        return "vv --progress"
    if VERBOSITY & 64:
        return "vvv --progress"
    if VERBOSITY >= 1:
        return " --progress"
    return ""


#!/usr/bin/env python3
import os, sys, argparse, subprocess, time, traceback, json, getpass
#
# the BDFL does not admire scripts which are also importable modules
# well, frack him -- this is how we roll
#
#so absolute imports work in script mode, we need to import from the parent folder
opath = os.path.abspath(".")
abspath = os.path.abspath(__file__)
# print ("BURST PATHS:", opath, abspath)
abspath = abspath[:abspath.rfind('/') + 1]
os.chdir(abspath)
abspath = os.path.abspath("..")
sys.path.insert(0, abspath)

from burst.lcloud import *
from burst.runrun import run
from burst.version import version
from burst.verbos import set_verbosity, get_verbosity, vprint, vvprint, v0print, get_piper, get_rsync_v, get_dockrunflags

os.chdir(opath)

DEFAULT_IMAGE = "burst_image" #FIXME: should be unique to folder structure

install_burst_sh = "sudo bash -c 'rm -fr /var/lib/dpkg/lock*" \
                   " /var/cache/apt/archives/lock /var/lib/apt/lists/lock;" \
                   "sudo systemctl stop apt-daily* ; " \
                   "apt-get -y update; " \
                   "apt-get -y install python3-pip; " \
                   "python3 -m pip install --upgrade pip; " \
                   "python3 -m pip install easydict apache-libcloud; " \
                   "rm -fr burst; " \
                   "git clone -b monitor https://github.com/burstable-ai/burst'"      #for reals

                # "git clone -b shutdown_39 https://github.com/danx0r/burst'"  # for testing


def do_ssh(url, cmd):
    ssh_cmd = f'ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error {url} ' \
          f'{cmd}'
    vvprint (ssh_cmd)
    os.system(ssh_cmd)


def ssh_tunnel(url, sshuser, ports, dockerdport):
    # set up ssh tunnel mapping docker socket, ports
    host_port_args = []
    docker_port_args = ""
    if ports:
        for pa in ports:
            if ':' in pa:
                local_port, remote_port = pa.split(':')
            else:
                remote_port = local_port = pa
            docker_port_args += " -p {0}:{0}".format(remote_port)
            host_port_args.append("-L {0}:localhost:{1}".format(local_port, remote_port))
    ssh_args = ["ssh", "-o StrictHostKeyChecking=no", "-o UserKnownHostsFile=/dev/null",
                "-o ExitOnForwardFailure=yes",
                "-o LogLevel=error", "-NL", "{0}:/var/run/docker.sock".format(dockerdport),
                "{0}@{1}".format(sshuser, url)]
    for arg in host_port_args:
        ssh_args.insert(3, arg)
    vvprint(ssh_args)
    tunnel = subprocess.Popen(ssh_args)
    vvprint("TUNNEL:", tunnel)
    time.sleep(2)
    return tunnel, docker_port_args


def burst(args, sshuser=None, url=None, uuid=None, burst_user=None, gpu=False, ports=None, stop=False,
          image=None, size=None, pubkey=None, dockerfile="Dockerfile",
          cloudmap="", dockerdport=2376, bgd=False, sync_only=False, conf=None):
    error = None
    tunnel = None
    try:
        if not os.path.exists(dockerfile):
            raise Exception("Dockerfile not found")
        if not os.path.exists(".dockerignore"):
            raise Exception("""

.dockerignore file not found. Burst requires a .dockerignore to avoid sending excess data to docker build.
Because the working directory is rsync'd to the remote host, you typically only need to send the Dockerfile
and files that are referred to (such as requirements.txt) to the build daemon.

#Template .dockerignore
#Default to ignoring everything:
**
#exceptions (These will be sent to the docker daemon for building):
!/Dockerfile*
!requirements.txt
""")

        #if url specified, split into user & IP
        if url:
            if not sshuser:
                sshuser, url = url.split('@')

        #launch, restart, or reconnect to node
        node = None

        #unless running --local:
        if url or uuid or burst_user:

            #if server does not exist, launch a fresh one
            fresh = False
            restart = False
            node = get_server(url=url, uuid=uuid, name=burst_user, conf=conf)
            if burst_user and not node:
                node = launch_server(burst_user, pubkey=pubkey, size=size, image=image, conf=conf, user=sshuser, gpu=gpu)
                fresh = True
                restart = True
            if node:

                #if stopped, restart
                if node.state.lower() != "running":
                    restart = True
                    vprint ("Starting server")
                    node = start_server(node)

                #by now we must have a public IP address
                url = node.public_ips[0]

                #wait for ssh daemon to be ready
                vprint ("Waiting for sshd")
                cmd = ["ssh", "-o StrictHostKeyChecking=no", "-o UserKnownHostsFile=/dev/null", "-o LogLevel=error", "{0}@{1}".format(sshuser, url), "echo", "'sshd responding'"]
                vvprint(cmd)
                good = False
                for z in range(10, -1, -1):
                    ret = run(cmd, timeout=15)
                    if ret[0].strip()[-15:]=='sshd responding':
                        good = True
                        break
                    vprint ("still waiting on sshd (this can take a while) -- will try %d more times" % z)
                    if z:
                        time.sleep(5)
                if not good:
                    raise Exception("error in ssh call: %s" % ret[0].strip())
                vvprint ("SSH returns -->%s|%s<--" % ret)
            else:
                raise Exception("Error: node not found")

        docker_port_args = ""

        #we have a url unless running --local:
        if url:

            #if just launched, install docker
            if fresh:
                vprint("Configuring Docker")
                # 'sudo apt-get -y update; sudo apt-get -y install docker.io; ' \ #images have docker installed
                cmd = 'ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error {0}@{1} ' \
                      '"sudo usermod -a -G docker ubuntu; ' \
                      'sudo systemctl unmask docker; sudo service docker start"'.format(sshuser, url)
                vvprint(cmd)
                os.system(cmd)

            vprint ("Connecting through ssh")
            tunnel, docker_port_args = ssh_tunnel(url, sshuser, ports, dockerdport)

            #path = absolute working directory on host
            relpath = os.path.abspath('.')[len(os.path.expanduser('~')):]
            relpath = "/_BURST" +  relpath.replace('/', '_') #I can exlain
            locpath = os.path.abspath('.')
            path = "/home/{0}{1}".format(sshuser, relpath)

            if not sync_only:
                # part of check to see if docker is installed and running
                remote = "-H localhost:%s" % dockerdport
                cmd = ["docker", "{0}".format(remote), "ps", "--format", '{{json .}}']
                vvprint(cmd)
                out, err = run(cmd)
                vvprint("PS returns:", out)
                running = len([x for x in out.strip().split("\n") if x])
                if running:
                    raise Exception("docker process already running -- burst does not support multiple processes")

                #prepare to build docker container
                vprint ("Removing topmost layer")        #to avoid running stale image
                cmd = ["docker", "{0}".format(remote), "rmi", "--no-prune", DEFAULT_IMAGE]
                vvprint (cmd)
                out, err = run(cmd)
                if "no such image" in out.lower():
                    out = "Creating new burst_image"
                vvprint (out)

            size, image = fix_size_and_image(size, image)
            if size and size != get_server_size(node):                      #FIXME
                raise Exception("Cannot change size (instance type) or gpu status -- need to re-launch")

            # get_server_image is broken, need to prompt better here
            # if image and image != get_server_image(node):
            # if image and image != get_server_image(node):
            #     raise Exception("FIXME: cannot change host image -- need to terminate & re-launch server")

            vprint ("burst: name %s size %s image %s url %s" % (node.name, size, image, url))

            #if using cloud storage (s3 etc), set up config & auth for rclone
            if cloudmap:
                if remote:
                    stor = get_config()['storage']
                    if stor['provider'] == 'GCS':
                        #create a keyfile & point to it
                        srvacctf = ".rclone_key_%s.json" % stor['settings']['private_key']['private_key_id']
                        f = open(srvacctf, 'w')
                        json.dump(stor['settings']['private_key'], f)
                        f.close()
                        stor['settings']['service_account_file'] = srvacctf

                    # build  & save rclone.conf
                    s = f"[{stor['config']}]\n"
                    for k, v in stor.items():
                        if k != 'settings':
                            s += f"{k} = {v}\n"
                    for k, v in stor['settings'].items():
                        s += f"{k} = {v}\n"
                    f = open(".rclone.conf", 'w')
                    f.write(s)
                    f.close()

            rsync_ignore_path = os.path.abspath("./.burstignore")
            if not sync_only:       #sync_only means from remote to local
                #sync local working data to host
                if not os.path.exists(rsync_ignore_path):
                    vprint("creating empty .burstignore")
                    os.system("touch .burstignore")
                cmd = 'rsync -rltzu{4} --exclude-from {5} -e "ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error" {0}/. {3}@{1}:{2}/'.format(locpath,
                                            url, path, sshuser, get_rsync_v(), rsync_ignore_path)
                vprint ("Synchronizing project folders")
                vvprint (cmd)
                os.system(cmd)

            if get_config().provider == 'GCE':
                # sync service acct creds (for shutdown)
                cmd = 'rsync -rltzu{4} --relative -e "ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error" {0}/./.burst/{5} {3}@{1}:{2}/'.format(os.path.expanduser('~'),
                                        url, path, sshuser, get_rsync_v(), get_config().raw_secret)
                vprint("Synchronizing credentials for shutdown")
                vvprint (cmd)
                os.system(cmd)

            #if fresh launch, clone burst locally for monitor
            if fresh:
                vprint ("Installing burst on server")
                vvprint("Delay for apt-get to settle")
                time.sleep(30)      #trust me this helps
                vvprint("Delay done")
                do_ssh(f"{sshuser}@{url}", '"%s"' % install_burst_sh)       #notable quoteables

            if restart:
                vprint ("Starting monitor process for shutdown++")
                #run monitor (in detached screen) to check if user's burst OR rsync is still running
                conf = get_config()
                if conf.provider == "GCE":
                    secret = ".burst/" + conf.raw_secret
                else:
                    secret = conf.secret
                proj = ('--project ' + conf.project) if conf.project else ''
                cmd = f"screen -md bash -c 'cd {path}; /usr/bin/python3 ~/burst/burst/monitor/monitor.py" \
                      f" --ip {url} --access {conf.access} --provider {conf.provider}" \
                      f" --secret={secret} --region {conf.region} {proj}'"
                vvprint (cmd)
                do_ssh(f"{sshuser}@{url}", '"%s"' % cmd)

        else:
            vprint ("burst: running locally")
            remote = ""
            path = os.path.abspath('.')

        if not sync_only:
            #actually build container -- for reals
            vprint ("Building docker container")
            cmd = "docker {1} build . --file {2} -t {0} {3}".format(DEFAULT_IMAGE, remote, dockerfile, get_piper())
            vvprint (cmd)
            os.system(cmd)

            #build argument list -- re-quote if whitespace
            # args = " ".join(args)
            s = ""
            for a in args:
                a = a.strip()
                if " " in a:
                    if '"' in a:
                        s += f"'{a}' "
                    else:
                        s += f'"{a}" '
                else:
                    s += f"{a} "
            args = s.rstrip()
            # print ("FINAL args:", args)
            # exit()

            if gpu:
                gpu_args = "--gpus all"
            else:
                gpu_args = ""

            #if mounting storage, add arguments & insert commands before (to mount) and after (to unmount) user-specified args
            cloud_args = ""
            if cloudmap:
                cloud, host = cloudmap.split(":")
                args = f"bash -c 'mkdir -p {host}; rclone mount --vfs-cache-mode writes --vfs-write-back 0 --config .rclone.conf {cloud}: {host} & sleep 3; {args}; umount {host}'"
                cloud_args = " --privileged"

            vprint ("Running docker container")
            background_args = "-td" if bgd else "-ti"
            cmd = f"docker {remote} run {gpu_args} {docker_port_args} --rm {background_args} --label ai.burstable.shutdown={stop} " \
                  f"-v {path}:/home/burst/work {cloud_args} {DEFAULT_IMAGE} {args}"

            #run main task
            vvprint (cmd)
            vprint ("")
            v0print ("---------------------OUTPUT-----------------------")
            sys.stdout.flush()
            if bgd:
                cmd = cmd.split()
                docker_container, err = run(cmd)
                print ("Running in background mode. Container =", docker_container[:11])
            else:
                os.system(cmd)
            sys.stdout.flush()
            v0print ("----------------------END-------------------------")
            sys.stdout.flush()

        #sync data on host back to local
        if url and not bgd:
            vprint ("Synchronizing folders")
            cmd = "rsync -rltzu{4} --exclude-from {5} -e 'ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error' '{3}@{1}:{2}/.' {0}/".format(locpath,
                                        url, path, sshuser, get_rsync_v(), rsync_ignore_path)
            vvprint (cmd)
            os.system(cmd)

    except Exception as ex:
        if get_verbosity() & 64:
            v0print ("--------------------------------")
            traceback.print_exc()
            v0print ("--------------------------------")
        else:
            print ()
        print (ex)
        error = "Exception"
    if tunnel:
        tunnel.kill()
    return error



import os
import configparser

def get_aws_creds(aws_path='~/.aws'):

    config_parser = configparser.ConfigParser()

    aws_abspath = os.path.expanduser(aws_path)
    config_parser.read(f'{aws_abspath}/credentials')
    config_parser.read(f'{aws_abspath}/config')

    index_map = {}
    print('Select AWS credentials to use:')
    index = -1
    for index, section in enumerate(config_parser.sections()):
        print(f'{index+1}: use "{section}" profile from {aws_abspath}')
        index_map[str(index+1)] = section

    env_access_key = os.environ.get('AWS_ACCESS_KEY_ID')
    env_secret = os.environ.get('AWS_SECRET_ACCESS_KEY')
    env_region = os.environ.get('AWS_DEFAULT_REGION')
    env_available = env_access_key is not None and env_secret is not None and env_region is not None

    if env_available:
        index += 1
        print(f'{index+1}: credentials in enviornment variables')

    print(f'{index+2}: Manually enter credentials')

    selected_index = input('\n1/2> ')

    if selected_index and selected_index in index_map:
        section = config_parser[index_map[selected_index]]
        access_key = section['aws_access_key_id']
        secret = section['aws_secret_access_key']
        region = section.get('region')
    elif env_available and selected_index == str(index+1):
        access_key = env_access_key
        secret = env_secret
        region = env_region
    elif selected_index == str(index+2):
        access_key = input('Access Key> ')
        secret = input('Secret Access Key> ')
        region = input('Region> ')
    else:
        raise Exception('Inavlid Selection')

    if region is None:
        print('This profile has no region')
        if 'region' in dict(config_parser).get('default'):
            default_region = config_parser['default']['region']
            choice = input(f'The "default" has region of {default_region}.  Should we use this? (y/n)> ')
            if choice.lower() == 'y':
                region = default_region
        if region is None:
            region = input('Please enter the region> ')

    return access_key, secret, region


if __name__ == '__main__':
    print(get_aws_creds())  # used for testing



def main_menu():

    input_map = {'1': 'compute', '2': 'storage', '3': 'default', '4': 'exit'}

    print('\nSelect from the options below:')
    print('  1. Set up or remove a compute service')
    print('  2. Set up or remove a storage service')
    print('  3. Change settings for compute or storage service (i.e. defaults)')
    print('  4. Exit configuration')
    selection = input(f"{'/'.join(input_map)}> ")

    if selection not in input_map:
        raise Exception("Invalid selection")

    return input_map[selection]


def main_service_menu(service):

    input_map = {'1': 'add', '2': 'remove'}

    print('\nDo you want to:')
    print(f'  1. Set up a new {service} service')
    print(f'  2. Remove a {service} service')

    selection = input(f"{'/'.join(input_map)}> ")

    if selection not in input_map:
        raise Exception("Invalid selection")

    return input_map[selection]


def default_service_menu():

    input_map = {'1': 'compute', '2': 'storage'}

    print('\nDo you want to:')
    print(f'  1. Change default compute service')
    print(f'  2. Change default storage service')
    selection = input(f"{'/'.join(input_map)}> ")

    if selection not in input_map:
        raise Exception("Invalid selection")

    return input_map[selection]
